/* error.c

	Error reporting and termination 
*/

#include "vpt.h"
#include "error.h"
#include "codes.h"
#include <signal.h>

static FLAG errcont;

static Codes *quitstr;

// static function prototypes

static void doquitcodes(void);


/*	============================================================
	Send/define the codes to send at the end of a print run.
	
	This will usually be a printer init. string.
*/

void
setqscodes(void)
{
	skipequal();
	
	setcodes(&quitstr,"quit string");
}

static void
doquitcodes(void)
{
	docodes(quitstr);
	flushoutput();
}


/*	============================================================
	Universal termination routine
*/

void
terminate(char *p, ...)
{
	va_list ap;
	
	errcont=printoff=0;
	
	fprintf(stderr, "VPRINT Fatal Error: ");
	if(p)
	{
		va_start(ap, p);
		vfprintf(stderr, p, ap);
		va_end(ap);
		fprintf(stderr, "\n");
		errlineout();
		maclineout();
	}
	fprintf(stderr, "\nFILE: %s, LINE: %d\n", get_current_filename(),
		get_current_linenumber());
	doquitcodes();
	exit(1);
}

void
error(char *p, ...)
{
	va_list ap;
	
	fprintf(stderr, "VPRINT Error: ");
	if(p)
	{
		va_start(ap, p);
		vfprintf(stderr, p, ap);
		va_end(ap);
	}
	fprintf(stderr, "\nFILE: %s, LINE: %d\n", get_current_filename(),
		get_current_linenumber());
	
	errlineout();
	maclineout();
	
	if((errcont || printoff) && jumpset) longjmp(jumpbuffer, 1);
	else
	{
		numcols=0;
		doquitcodes();
		exit(1);
	}
}

/* abort with user error message */

void
userr(void)
{
	char buff[200];
	
	skipequal();
	getstring(buff, sizeof(buff)-2, "user error message");
	terminate("USER-%s", buff);
}

/*	============================================================
	Intercept handler. Used to ensure file cleanup when keyboard-
	abort hit. All signals other than 2 and 3 are ignored....
*/

void
icepthand(int signal)
{
	if(signal==SIGQUIT || signal==SIGINT)
	{
		doquitcodes();
		fprintf(stderr, "\nAbort signal received, VPRINT terminated\n");
		exit(signal);
	}
}

/*	============================================================
	This is after an 'qu' is encountered. The column buffer is flushed,
	but no newpage is done. This is also called after EOF, but the main
	loop first starts a new page.
*/

void
quit(void)
{
	if(numcols) flushcols();
	doquitcodes();
	exit(0);
}

/*	============================================================
	Set the error continue flag
*/

void
seterrflag(int f)
{
	errcont=f;
}

/*	============================================================
	Usage, called from command line interperter
*/

void
usage(void)
{
	static char *msg[]=
	{
		"Usage: Vpt [<options>] <file> {<file> <options>}",
		"Function: Print formatted copy of file",
		"Options:",
		"	-e		Continue after errors",
		"	-f <text>	Initial command line",
		"	-i <file>	Specify init file",
		"	-p n1,n2	Page range to print",
		"	-p O/E		Print Odd/Even pages",
		"	-s <file>	Spool each page to a file",
		"	-o <path>	Set output file to <path>",
		NULL
	};
	
	char **p=msg;
	
	while(*p) fprintf(stderr, "%s\n", *p++);
	exit(0);
}
