/* format.c

   Stuff the buffer with one line of text. We have to handle things
   like footnotes and marginal notes here.
   
   Justify, line formatting routines.
*/

#include "vpt.h"
#include "att_table.h"

/* Line format for text, headers and notes.
   0 for even, 1 for odd pages
*/

static struct fmt_struct
{
	char	f_text[2],
			f_head[2],
			f_note[2],
			f_mnote[2];
} linefmts;

static u_int tabcolumns=MS_DEFAULT*4;	/* set tab cols to 4/10" inch */
static u_int interspacing;			 	/* intercharacter spacing value */
static FLAG jcline,						/* set==justify CR lines */
			linead;						/* set==use CR terminated lines */
			
static u_char justmode;					/* justify mode  */
static u_char padcharmode;				/* fill prop chars */
static u_char padcharsize;				/* size to fill prop. to */

// function prototypes

static int widetab(KCHAR *, KCHAR *, int, int, Att_table *);
static int centerline(KCHAR *, KCHAR *, int, int, Att_table *);
static int justright(KCHAR *, KCHAR *, int, int, Att_table *);
static int justfull(KCHAR *, KCHAR *, int, int, Att_table *);
static void insertchar(KCHAR *, KCHAR *, KCHAR);
static void bufferfull(void);
static int shortenline(KCHAR *, KCHAR *, int, int, Att_table *);
static void update_at(KCHAR *, KCHAR *, Att_table *, Att_table *);
static int padtab(KCHAR *, KCHAR *, int, int, Att_table *);
static KCHAR *toendofbuff(KCHAR *);
static void pushatoffs(char *, int);


/* ======================================================================
	Set proportional character padding size
*/

void
setpadcharsize(void)
{
	skipequal();
	padcharsize=getAval();
}

void
setcharpadding(KCHAR c)
{
	if(c==INT_PAD_ON) padcharmode=1;
	else if(c==INT_PAD_OFF) padcharmode=0;
}

/* ============================================================
   Set the tab column size.
*/

void
settabcolumn(void)
{
	skipequal();
	tabcolumns=getval(1);
}

/* ============================================================
   Set the intercharacter spacing value
*/

void
setispacing(void)
{
	skipequal();
	interspacing=getBval(0);
}

/* ============================================================
   Set the line format  LF[n,h][o,e]=L,F,C,J
*/

void
setlineformat(void)
{
	char ftype, foe;
	KCHAR opt;
	char *p;
	
	whichmarg(&ftype, &foe);
	
	skipequal();
	opt=f_toupper(getkey());
	
	if(index("LRJC", opt)==0) error("Illegal line format setting");
	
	switch(ftype)
	{
		case 'n':
			p=linefmts.f_note;
			break;
			
		case 'h':
			p=linefmts.f_head;
			break;
			
		case 'm':
			p=linefmts.f_mnote;
			break;
			
		default:
			p=linefmts.f_text;
			break;
	}
	
	switch(foe)
	{
		case 'e':
			p[0]=opt;
			break;
			
		case 'o':
			p[1]=opt;
			break;
			
		default:
			p[0]=p[1]=opt;
	}
}

/* ============================================================
   Expand wide tabs in the buffer
*/

static int
widetab(KCHAR *buff, KCHAR *buffend, int currlen, int maxlen, Att_table *startatts)
{
	KCHAR *p, c, filler;
	int fillsize, smallsize;
	Att_table curratts;
	
	memcpy(&curratts, startatts, sizeof(Att_table));
	smallsize=smallspacesize();
	
	p=buff;
	for( ; ; )
	{
		if(is_font(c=*p) || is_attron(c) || is_attroff(c))
		{
			update_at(buff, p, &curratts, startatts);
			p++;
		}
		else if(c==INT_WIDETAB)
		{
			filler=tabfill;
			fillsize=getcsize(filler, &curratts);
			if(fillsize+currlen>maxlen)
			{
				filler=INT_SSPACE;
				fillsize=smallsize;
				if(!smallsize || currlen+fillsize>maxlen) break;
			}
			insertchar(p, buffend, filler);  /* insert & check for overflow */
			currlen+=fillsize;
			p+=2;  /* skip filler and token */
		}
		else if(c==0)	   /* end of buffer, start over */
		{
			p=buff;
			memcpy(&curratts, startatts, sizeof(Att_table));
		}
		else p++;
	}
	return currlen;
}

static int
centerline(KCHAR *buff, KCHAR *buffend, int currlen, int maxlen, Att_table *startatts)
{
	return justright(buff, buffend, currlen,
		maxlen-((maxlen-currlen)/2), startatts);
}

static int
justright(KCHAR *buff, KCHAR *buffend, int currlen, int maxlen, Att_table *startatts)
{
	char c;
	
	insertchar(buff, buffend, INT_WIDETAB);
	buffend++;
	c=tabfill;
	tabfill=' ';
	currlen=widetab(buff, buffend, currlen, maxlen, startatts);
	tabfill=c;
	return currlen;
}

static int
justfull(KCHAR *buff, KCHAR *buffend, int currlen, int maxlen, Att_table *startatts)
{
	KCHAR *p, c, filler;
	int s_size=smallspacesize();
	char jm=justmode;
	FLAG changes, jflag;
	int t;
	int fillsize;
	
	Att_table currtable;
	
	if(!s_size) jm=0;  /* have to use justify mode 0 if no small space */
	
	for(t=0, p=buff; *p++; t++);  /* count chars in buff */
	
	/*	'jflag' is used to toggle between jm3 and jm1 for alternate passes
		when jm is set to 2. This will force alternate lines to fill either
		all points, or only at spaces. The loop will never see a value of
		2 for jm.
	*/
	
	if(jm==2) jm=jflag=1;
	else jflag=0;
	
	
	p=buff+rnd(t);
	update_at(buff, p, &currtable, startatts);
	changes=1;
	if(jm==0) filler=' ';
	else
	{
		fillsize=s_size;
		filler=INT_SSPACE;
	}
	
	for( ; ; )
	{
		c=*p++;
		
		if(!c)
		{					 /* reached the end of the buffer */
			if(!changes)
			{				/* if no changes, bump from 0 to 1 */
				if(jm==0 && s_size)
				{
					jm=1;
					filler=INT_SSPACE;
					fillsize=s_size;
				}
				else break;				  /* or exit. */
			}
			p=buff;				 /* start over at line start */
			update_at(buff, p, &currtable, startatts);
			changes=0;
			if(jflag) jm=jflag=(jflag==1) ? 3 : 1;
			continue;
		}
		
		if(istoken(c))
		{
			if(is_font(c) || is_attron(c) || is_attroff(c))
			{
				update_at(buff, p, &currtable, startatts);
			}
			continue;
		}
		
		if(jm<=1 && c==SPACE && *p!=SPACE)
		{
			if(jm==0)
			{
				fillsize=getcsize(filler, &currtable);
				if(fillsize+currlen>maxlen) continue;
			}
			if(jm==1 && fillsize+currlen>maxlen) break;
			
			insertchar(p++, buffend, filler); /* insert & check overflow */
			currlen+=fillsize;
			changes=1;
			continue;
		}
		
		if(jm==3)
		{
//			if(hardspace && (c==hardspace || *p==hardspace)) continue;
			if(*p==INT_HSPACE) continue;
			if(fillsize+currlen>maxlen) break;
			insertchar(p++, buffend, filler);
			currlen+=fillsize;
			changes=1;
		}
	}
	return currlen;
}


/* Insert one character into the line buffer and check for buffer overflow */

static void
insertchar(KCHAR *buff, KCHAR *buffend, KCHAR c)
{
	KCHAR c1, c2;
	
	c1=c;
	
	do
	{
		c2=*buff;
		*buff++=c1;
	}
	while((c1=c2));
	
	*buff=0;
	if(buff>=buffend) bufferfull();
}

static void
bufferfull(void)
{
	terminate("Format buffer overflow");
}


/* ============================================================
   Shorten the formatted line for word wrap
*/

static int
shortenline(KCHAR *buff, KCHAR *dataend, int currlen, int maxlen, Att_table *startatts)
{
	int minlen=currlen/2;
	KCHAR c, *firstspace, *p;
	int t;
	
	Att_table curratts;
	update_at(buff, dataend, &curratts, startatts);
	
	dataend--;  /* point to last char, not NULL */
	if(dataend<=buff) goto EXIT;
	
	/* strip off trailing microspaces */   
	
	while(*dataend==INT_SSPACE && dataend>buff)
	{
		dataend--;
		currlen-=getcsize(INT_SSPACE, &curratts);
	}
	
	/*	trailing hypen check. If the last char in the buffer is
		a single '-' then get the next one to see if we have split
		a hyphen. If the next char is a '-' then add it to the buffer.
	*/
	
	if(*dataend=='-' && *(dataend-1)!='-')
	{
		c=getkey();
		if(c=='-')
		{
			*(++dataend)=c;
			currlen+=getcsize(c, &curratts);
		}
		else returnkey(c);
	}
	
	/*	If the line contains no spaces, hypens or ghost hyphens then
		just leave it alone.
	*/
	
	for(p=buff; ; )
	{
		c=*p++;
		if(c=='-' || c==SPACE || c==INT_GH) break;
		if(p==dataend) goto EXIT;
	}
	
	/*	Now we have to backup to a point in the buffer which is
		either a word break, a hypen, a ghost hypen or if the new
		line size is less than 1/2 the original line size and there
		are no more spaces to backup to in the line.
	*/
	
	for(firstspace=buff; firstspace<dataend; firstspace++)
	{
		if(*firstspace==SPACE) break;
	}
	
	for( ; ; )
	{
		if((currlen<=minlen && dataend<=firstspace) ||
			 dataend<=buff) goto EXIT;
		
		c=*dataend;
		
		/*	If we are at a '-' or a ' ' it is okay to break the line. Note
			that we also check to make sure that the line is short enough!
			Note that trailing spaces are discarded outside of the loop.
		*/
		
		if(currlen<=maxlen && (c==' ' || (c=='-' && f_isalnum(*(dataend+1)))))
			break;
		
		if(istoken(c) && (is_font(c) || is_attroff(c) || is_attron(c)))
		{
			update_at(buff, dataend, &curratts, startatts);
		}
		
		/*	if we find a ghost hyphen we first check to see if adding
			a real hyphen at this point will make the linesize okay. If
			true, we insert a hyphen and exit; otherwise we discard
			the ghost and continue
		*/
		
		if(c==INT_GH)
		{
			if(currlen+(t=getcsize('-', &curratts))<=maxlen)
			{
				*dataend='-';
				currlen+=t;
				break;
			}
			else
			{
				dataend--;  /* discard ghost */
				continue;
			}
		}
		
		returnkey(c);
		dataend--;
		currlen-=getcsize(c, &curratts);
		
		/*	Hyhpen check, if the character just stripped was a '-' we check
			to see if the prior one is a '-' too. If it is we've got a hypen
			so we return the first half too. If the hypen is attached to the
			prior word then all is okay since the break will occur at the 			start of the word.
		*/
		
		if(c=='-' && *dataend==c)
		{
			returnkey(c);
			currlen-=getcsize(c, &curratts);
			dataend--;
		}
	}
	
	/* strip off the trailing spaces from remaining line */
	
	while(dataend>buff && ((c=*dataend)==' ' || c==INT_SSPACE))
	{
		dataend--;
		currlen-=getcsize(c, &curratts);
	}
	
EXIT:;
	*(dataend+1)=0;  /* null terminate the buffer */
	
	return currlen;
}


/*	============================================================
   	Update the attribute table to reflect the current position
	in the line. This is used by justify and shorten to make sure
	the character size being accessed matches with the font/attr
	selection at the current point in the line.
*/

static void
update_at(KCHAR *buff, KCHAR *curpoint, Att_table *currtbl, Att_table *starttbl)
{
	KCHAR c;
	
	memcpy(currtbl, starttbl, sizeof(Att_table));
	
	while(buff<=curpoint)
	{
		c=*buff++;
		
		if(is_font(c)) currtbl->cfont=c-INT_FONT;
		
		else if(is_attron(c))
		{
			c-=INT_ATTRON;
			currtbl->aselect[c]=1;
			if(getasize(c)) currtbl->sizemod=getasize(c);
		}
		
		else if(is_attroff(c))
		{
			c-=INT_ATTROFF;
			currtbl->aselect[c]=0;
			if(getasize(c)) currtbl->sizemod=1;
		}
	}
}


static int
padtab(KCHAR *p, KCHAR *buffend, int size, int currlen, Att_table *curratts)
{
	KCHAR filler;
	int s;
	
	if(!(filler=tabfill)) filler=' ';
	
	s=getcsize(filler, curratts);
	while((size-s)>=0)
	{
		*p++=filler;
		size-=s;
		currlen+=s;
		if(p>=buffend) bufferfull();
	}
	if((s=getcsize(INT_SSPACE, curratts)))
	{
		while((size-s)>=0)
		{
			*p++=INT_SSPACE;
			size-=s;
			currlen+=s;
			if(p>=buffend) bufferfull();
		}
	}
	*p=0;
	return currlen;
}

static KCHAR *
toendofbuff(KCHAR *buff)
{
	KCHAR *p=buff;
	
	while(*p) p++;
	return p;
}

/*	 ============================================================
	 Fill the line buffer with one line of text for printing.
*/

static void
pushatoffs(char *list, int size)
{
	int t;
	for(t=0; t<size; t++)
	{
		if(list[t]) returnkey(INT_ATTROFF+t);
	}
}

/* Fill buffer.
	buff		ptr to buffer
	buffend		ptr to end of buffer
	maxlen		line length needed in MS
	indentsize
	centerid
	startatts
	mode		0==main text, 1==header
*/

int
fillbuff(KCHAR *buff, KCHAR *buffend, int maxlen, int indentsize, int centerid, Att_table *startatts, int mode)
{
	int currlen=0;	 /* number of chars in buffer */
	KCHAR *p=buff;
	KCHAR c;
	int t;
	KCHAR *startjust=buff;
	int origmaxlen=maxlen;
	KCHAR c1;
	int ssize;
	
	Att_table curratts;
	
	char wordatts[NUM_ATTRIBUTES];
	
	FLAG wactive=0, 	/* 1=doing a word attribute			   */
	iddone=0,			/* 1=delayed indent done is this line  */
	ateol=0;			/* 1==CR terminated line			   */
	
	int word_ul=getregnum('v');
	
	memset(wordatts, 0, sizeof(wordatts));  /* no word atts active */
	
	memcpy(&curratts, startatts, sizeof(Att_table));
	
	if(centerid) maxlen=(maxlen-centerid)/2;
	
	for( ; ; )
	{
		c=getkey();
		
		if(c==CR && linead && currlen && mode==FMT_TEXT)
		{
			c1=getkey();
			
			returnkey( (c1==CR) ? INT_CR : c1);
			
			if(c1!=CR && c1!=SPACE && !f_ispunct(c1)) c=SPACE;
		}
		
		if(c==INT_CR) c=CR;
//		if(ghosthy && c==ghosthy) c=INT_GH;  /* ghost hypens to internal */
		if(!c) c=CR;
		
		if(c==CR)	 /* end of line */
		{
			if(wactive)
			{
				returnkey(c);
				pushatoffs(wordatts, sizeof(wordatts));
				wactive=0;
				continue;
			}
			ateol++;
			returnkey(c);
			break;
		}
		
		/*	special process for attribute 'v' which is word underlining.
			If selected we surround change each space to a v-<space>v+.
		*/
		
		else if((c==SPACE || is_tab(c)) && curratts.aselect[word_ul])
		{
			returnkey(INT_ATTRON+word_ul);
			returnkey(c);
			returnkey(INT_ATTROFF+word_ul);
			continue;
		}
		
		/* special process for word attributes (\a*). All word attributes
		   are saved on a stack when turned on. When a space if found they
		   are turned off by inserted a \a- into the input stream
		*/
		
		else if(wactive && (c==SPACE || is_tab(c) ||
			is_tabset(c) || c==INT_COLTAB))
		{
			returnkey(c);
			pushatoffs(wordatts, sizeof(wordatts));
			wactive=0;
			continue;
		}
		
		/*	Start a footnote, suck the text out of the stream and save
			it. A new page, etc. will be started if nesc by extractnote().
		*/
		
		else if(c==INT_NOTE)
		{
			if(mode!=FMT_TEXT)
				error("Notes are only permitted in main text");
			
			extractnote();	/* get the footnote text & start new */
			continue;		/* page if necessary	*/
		}
		
		/* Toggle fonts and attributes */
		
		else if(is_font(c))
		{
			t=c-INT_FONT;
			if(curratts.cfont==t) continue;	  /* ignore dup font change */
			curratts.cfont=c-INT_FONT;
		}
		
		else if(c==INT_WORDATT || is_attron(c))
		{
			if(c==INT_WORDATT)
			{
				*p++=c;				 /* save word att in buff */
				c=getkey();			 /* get the attribute	  */
				wordatts[c-INT_ATTRON]++;	 /* set the word att flag */
				wactive=1;
			}
			t=c-INT_ATTRON;
			if(curratts.aselect[t]) continue;  /* skip duplicate on codes */
			curratts.aselect[t]++;
			if(getasize(t)) curratts.sizemod=getasize(t);
		}
		
		else if(is_attroff(c))
		{
			t=c-INT_ATTROFF;
			if(curratts.aselect[t]==0) continue; /* neven off, ignore */
			curratts.aselect[t]=0;
			if(getasize(t)) curratts.sizemod=1;
		}
		
		/*	Tab to a column tabstop. We just pad the line until the we're
			at a tab column.
		*/
		
		else if(c==INT_COLTAB)
		{
			t=currlen+indentsize;
			t=((t/tabcolumns)+1)*tabcolumns;
			if(t-indentsize > maxlen)
			{
				returnkey(CR);
				break;
			}
			else
			{
				t-=(currlen+indentsize);
				currlen=padtab(p, buffend, t, currlen, &curratts);
				p=toendofbuff(p);
			}
			startjust=p+1;
			continue;
		}
		
		/*	Tab to specified stop. We just pad out the line to make it the
			needed size. We also set 'startjust' to signal that no
			justification should occur before the tabstop.
		*/
		
		else if(is_tab(c))
		{
			currlen=padtab(p, buffend, gettabvalue(c-INT_TAB)-
				currlen-indentsize, currlen, &curratts);
			p=toendofbuff(p);
			startjust=p+1;
			continue;
		}
		
		/* Start a delayed indent. Again, like tabs, we set 'startjust' */
		
		else if(c==INT_DELAYID)
		{
			if(mode==FMT_TEXT)
			{
				if(iddone) del_dytab();
				add_dytab(currlen+(indentsize-getleftid()));
				iddone=1;
			}
			else if(mode==FMT_NOTE) setnoteid(currlen);
			else error("\\ID only permitted in main text and footnotes");
			
			startjust=p+1;
			continue;
		}
		
		/*	Set an embedded tabstop. We find the current line pos and use
			it to set the stop. We also set 'startjust' to signal that no
			justification should occur before the tabstop.
		*/
		
		else if(is_tabset(c))
		{
			if(mode!=FMT_TEXT) error("Tabs can only be set in main text");
			setdytab(c-INT_TABSET, currlen+indentsize);
			startjust=p+1;
			continue;
		}
		
		else if(c==INT_PAD_ON || c==INT_PAD_OFF)
		{
			setcharpadding(c);
			continue;
		}
		
		
		/* End of conditional processing. Stuff char/token into the buffer */
				
		*p++=c;
		if(p>=buffend) bufferfull();
		currlen+=getcsize(c, &curratts);
		
		/* add intercharacter spacing, done after the character */
		
		if(interspacing && (ssize=getcsize(INT_SSPACE, &curratts)))
		{
			for(t=interspacing; t; t--)
			{
				*p++=INT_SSPACE;
				if(p>=buffend) bufferfull();
				currlen+=ssize;
			}
		}
		
		/*	Pad a proportional char to the same size as something else.
			This is used mostly to nicely line up columns of numbers, etc.
			It is only valid when doing the main body.
			
			Note that we can't use padtab() to do this since it may insert
			SPACEs and that will cause wrap and justify problems.
			
			This is done after the character.
		*/
		
		if(padcharmode && mode==FMT_TEXT && 
			(ssize=getcsize(INT_SSPACE, &curratts)))
		{
			if(padcharsize==0) padcharsize='w';
			t=getcsize(padcharsize, &curratts)-getcsize(c, &curratts);
			while(t>0)
			{
				*p++=INT_SSPACE;
				if(p>=buffend) bufferfull();
				currlen+=ssize;
				t-=ssize;
			}
		}
		
		/*	Test to see if line if full. This is also where center ids are
			set, since all we did for center ids was to shorten the line size.
		*/
		
		if(currlen>=maxlen && !ateol)
		{
			if(centerid)
			{
				t=currlen;
				currlen=shortenline(buff, p, currlen, maxlen, startatts);
				p=toendofbuff(buff);
				if(startjust<p && linefmts.f_text[pageodd()]=='J')
					currlen=justfull(startjust, buffend, currlen, 
						maxlen, startatts);
				p=toendofbuff(buff);
				update_at(buff, p, &curratts, startatts);
				t=t-currlen;
				currlen=padtab(p, buffend, centerid+t, currlen, &curratts);
				p=toendofbuff(p);
				centerid=0;
				maxlen=origmaxlen;
				startjust=p+1;
			}
			else break;
		}
	}
	*p=0;
	
	if(mode==FMT_CBAR || mode==FMT_PREFIX || mode==FMT_SUFFIX) goto exit;
	
	if(!ateol)
	{
		currlen=shortenline(buff, p, currlen, maxlen, startatts);
		while((c=getkey())==SPACE);  /* we don't want next line to */
		returnkey(c);				 /* start with a space... */
	}
	
	/*	The line has been read into the buffer and has been word-wrapped.
		 Now justify it, etc.
		 
		 No matter what the format, we have to expand wide tabs.
		 Line format settings are ignored for change bars and marginal notes.
		 
		 First we check to see if this is worthwhile...if we're not printing
		 or if this is an empty line we just exit.
	*/
	
	if(!printpg || !currlen) goto exit;
	
	for(p=buff; *p; )
	{
		if(*p++==INT_WIDETAB)
		{
			currlen=widetab(buff, buffend, currlen, maxlen, startatts);
			goto exit;
		}
	}

	/* format the line according to the format flag */

	switch(f_toupper(getlineformat(mode, pageodd())))
	{
		case 'R':
			currlen=justright(buff, buffend, currlen, maxlen, startatts);
			break;
		
		case 'J':
			if(ateol && !jcline) break;  /* don't justify CR lines */
			for(p=buff; *(p++); );  /* move p to eol */
			if(startjust<p)
				currlen=justfull(startjust, buffend, 
					currlen, maxlen, startatts);
			break;
		
		case 'C':
			currlen=centerline(buff, buffend, currlen, maxlen, startatts);
			break;
		
		default:
			break;		/* should never happen! */			
	}
	
exit:
	for(p=buff; *p; p++)			 /* convert hardspaces in buffer */
	{
//		if(hardspace && *p==hardspace) *p=' ';	  /* to real spaces  */
		if(*p==INT_HSPACE) *p=' ';
		
	}
	return currlen;
}


/*	============================================================
	Return a pointer to the start of the line format vars and its
	size, needed by state.
*/

void
getfmtstate(int *s, char **p)
{
	*s=sizeof(struct fmt_struct);
	*p=(char *) &linefmts;
}

/*	============================================================
 	Return current line format for this page. Called from math and format
	
	Get the format for this line (LCJR). There is no format for change
	bars and line pre/suffixes.
*/

int
getlineformat(int mode, int pg)
{
	char c;
	
	switch(mode)
	{
		case FMT_TEXT:
			c=linefmts.f_text[pg];
			break;
		
		case FMT_NOTE:
			c=linefmts.f_note[pg];
			break;
		
		case FMT_HEAD:
			c=linefmts.f_head[pg];
			break;
		
		case FMT_MARG:
			c=linefmts.f_mnote[pg];
			break;
		
		default:
			c='L';
	}
	return f_tolower(c);
}

/*	======================================================================
	Flag setting routines
*/

void
set_jcline(void)
{
	jcline=getonoff();
}

void
set_linead(void)
{
	linead=getonoff();
}

void
set_justmode(void)
{
	skipequal();
	justmode=getBval(0);
}

int
get_justmode(void)
{
	return justmode;
}
/*	======================================================================
	Reset the tabcolumn size after a MS= command
*/


void
ms_tabcols(u_int n)
{
	tabcolumns=ms_adjust(tabcolumns, n);
	interspacing=ms_adjust(interspacing, n);
}
