/* macros.c */

#include "vpt.h"
#include <sys/stat.h>
#include <time.h>

#define MAXMACS 20

/*	This structure sets up a frame for a macro definition. These
	definitions are stored in a linked list. The data consists of a
	name, a null and the data. We also save the name of the file being
	processed and the current line number.
*/

typedef struct macptr
{
	struct macptr *next;	 /* ptr to next string */
	char *nameptr;
	u_char *dataptr;
}Macptr;

static Macptr *root, *last;

static u_char	*mac_current[MAXMACS],	/* current macro ptr	*/
		*macstartpos[MAXMACS];	/* start of macro ptr	*/
static char  	*cmacname[MAXMACS];	/* macro name ptr	*/
static u_int	macro_linenumber[MAXMACS];
static u_char 	*macro_filename[MAXMACS];

static void getmacname(char *buff, int maxlen);


/*	============================================================
	Parse off a macro name. We assume that the leading "$" or "@"
	has been skipped over already.
	
	The name is stored in the passed buffer in uppercase.
*/

static void
getmacname(char *buff, int maxlen)
{
	char *p=buff;
	KCHAR c;
	
	for( ; ; )
	{
		if((c=getkey())=='@' || c=='$')
		{
			*p=0;
			if(p==buff) error("A null macro name is illegal");
			break;
		}
		if(c<=' ' || c>=0x7f)
		{
			*p=0;
			error("Illegal character '$%02x' in macro name '%s'", c, buff);
		}
		*p++=(char)f_toupper(c);
		if(--maxlen<=0) error("Macro name too long");
	}
}


/* ============================================================
   Set the filename macros. Called from newfile() and getcinput().
   
   f - file to set up
   n - name of the file
*/

void
setfmacs(FILE *f, char *n)
{
	char datebuff[100];
	struct stat sbuff;
	
	add_macro_def("C_FILE", n);
	
	if(!isrbf(f)) sprintf(datebuff, "UNKNOWN");
	
	else
	{
		if(fstat(fileno(f), &sbuff)!=0)
			error("Can't read file descriptor for current file");
		strftime(datebuff, sizeof(datebuff), "%H:%M:%S %y/%m/%d",
			localtime(&sbuff.st_mtime));
	}
	add_macro_def("M_DATE", datebuff);
}


/* ============================================================
   Define a macro
   
   .$name$=<text>
   .@name@=<text>
*/

void
defmacro(void)
{
	char namebuff[50];
	u_char textbuff[500];
	
	/*	Get the name of the macro and the data. Both are buffered,
		errors if name or data is too long.
	*/
	
	getmacname(namebuff, arraysize(namebuff)-2);  /* parse off the name */
	skipequal();
	getstring(textbuff, arraysize(textbuff)-2, "macro definition");
	add_macro_def(namebuff, textbuff);
}

/*	Actually add/define macro. This is called from above and from
	newfile() to add internal macros $FILE$ $C_FILE$ $M_DATE$
*/


void
add_macro_def(char *name, u_char *text)
{
	Macptr *temp;
	int t;
	
	/*	Check if this macro is already defined. If it is we delete its data.
		Note that we also check the macros used stack to see if the macro
		is redefining itself, in which case we error out.
	*/
	
	for(temp=root; temp; temp=temp->next)
	{
		if((strcmp(temp->nameptr, name))==0)
		{
			for(t=0; t<MAXMACS; t++)
			{
				if(macstartpos[t]==temp->dataptr)
					error("Can't recursively redefine macro");
			}
			if(temp->dataptr) free(temp->dataptr); /* dump old definition */
			temp->dataptr=NULL;
			break;
		}
	}
	
	if(!temp)  /* define a new macro, set up name and struct memory */
	{
		if((temp=(Macptr *)malloc(sizeof(Macptr)))==NULL)
			terminate("No memory for macro storage");
		temp->dataptr=0;
		temp->nameptr=strdup(name);
		
		/* Set the forward links */
		
		temp->next=NULL;	/* new macros always to end of list	*/
		
		if(root==0) root=temp;	/* 1st one, set root			*/
		else last->next=temp;	/* else 2nd+ macro, add to list		*/
		
		last=temp;		/* set ptr to new entry			*/
	}
	
	/* add the text to the package */
	
	temp->dataptr=(u_char *)strdup(text);
}

/*	============================================================
	Expand a macro encountered in the input stream. This is called
	from getkey() in io.c.
   	
	Note that we set two pointers. The first (mac_current) is a pointer
	to the data, the second (macstartpos) is copy of the macro pointer.
	This pointer is checked when defining macros to see if we are
	attempting to redefine a macro being executed; it is also used by
	goto so find the start of the macro.
*/

void
expmac(void)
{
	char nambuff[50];
	Macptr *temp;
	
	getmacname(nambuff, arraysize(nambuff)-2);  /* get name of macro */
	
	/* look for the name in the linked list of macro pointers */
	
	for(temp=root; ; temp=temp->next)
	{
		if(!temp) error("Macro '%s' not defined", nambuff);
		if((strcmp(temp->nameptr, nambuff))==0) break;
	}
	
	/* TEMP points to the correct macro */
	
	addmacro(temp->dataptr);
	cmacname[maclevel-1]=strdup(nambuff);
	macro_linenumber[maclevel-1]=get_current_linenumber();
	macro_filename[maclevel-1]=get_current_filename();
}

/*	============================================================
	Add a macro to the stack for execution. The new macro level
	is returned to the caller.
*/


int
addmacro(u_char *p)
{
	if(maclevel>=MAXMACS) terminate("Macro stack overflow");
	mac_current[maclevel]=macstartpos[maclevel]=p;
	
	if(cmacname[maclevel])		  /* delete currently executing */
	{
		free(cmacname[maclevel]);	/* macro name if extant */
		cmacname[maclevel]=0;
	}
	return ++maclevel;
}


/*	============================================================
	Get a key from a macro. If no key (macro) pending return 0
*/

KCHAR
getcmac(void)
{
	KCHAR c=0;
	u_char *mptr;
	
	while(maclevel && !c)
	{
		mptr=mac_current[maclevel-1];		/* point to macro	*/
		c=*mptr++;				/* get char, update ptr	*/
		if(c==INT_TOKEN) c=0x100+(*mptr++);
		mac_current[maclevel-1]=mptr;		/* remember ptr for next time */
		if(!c || *mptr==0) killmacro(maclevel);	/* macro done, end it */
	}
	return c;
}

/*	============================================================
	End processing of the specified macro. This will only work for
	the currently executing macro AND the param passed has to match
	the current macro level. This is called by getcmac() to terminate
	a macro when a NULL is encountered, it is also called by routines
	which interupt normal input by starting a macro (fnotes, etc.) and
	by calling this they can be ensured that the macro is done.
*/

void
killmacro(int n)
{
	while(maclevel>=n)
	{
		maclevel--;
		macstartpos[maclevel]=0;
	}
}

/*	============================================================
	Display the current macro. Called from the error termination
	routine. This also cancels all pending macros in case we decide
	to continue after the error...
*/

void
maclineout(void)
{
	char *p=cmacname[maclevel-1];
	
	if(maclevel)
	{
		if(!p) p="$$internal$$";
		fprintf(stderr, "Macro '%s' (%s:%d)\n%s\n", p, macro_filename[maclevel-1],
			macro_linenumber[maclevel-1], macstartpos[maclevel-1]);
		maclevel=0;
	}
}


/*	============================================================
	Search for a label in the current macro. Just as in fill goto
	searches the label must start a dot line. Note that we check
	the macro from start to end, not from current pos as in the file
	check.
   	
   	If the label is found we set the current macro pointer to the
   	end of the label for continuation.
   	
   	Returns 1 - label found, 0 - not found
*/

int
gotomacro(char *label)
{
	KCHAR c=0;
	u_char *mptr;
	char *p;
	FLAG flag=0;
	
	if(maclevel)
	{
		for(mptr=macstartpos[maclevel-1], c=CR; *mptr; c=*mptr++)
		{
			if(c!=CR) continue;		// find a CR
			if(*mptr!=dot) continue;	// followed by a '.'
			while(*(++mptr)==' ');		// skip leading blanks
			if(*mptr!=':') continue;	// loop if not label start
			
			/* found a label, see if it matches target */
			
			p=label;
			mptr++;
			while(*mptr && f_tolower(*mptr)==f_tolower(*p))
			{
				mptr++;
				p++;
			}
			if(*p=='\0')
			{	  			/* did we reach end of target? */
				flag++;			/* yes, set flag for return	   */
				mac_current[maclevel-1]=mptr;	/* set up new macro pointer*/
				break;			/* leave for() loop */
			}
		}
	}
	return flag;
}

