/* numbers.c

   Number register commands
*/

#include "vpt.h"
#include <time.h>

#define NUMSTRMAX 40

typedef struct num_struct
{
	u_int	num_value,
			num_inc;
	char	num_fmt,
			num_pad,
			num_precise;
}Num_struct;

static Num_struct numbers[36];


static KCHAR numberstring[NUMSTRMAX];
static char numstringcount;


static Num_struct *pointnumreg(KCHAR n);
static void itoroman(char *p, int n);
static int itoeng(char *dest, int n);

/*	============================================================
	Return a pointer to the correct number register
*/

static Num_struct *
pointnumreg(KCHAR n)
{
	int t;
	
	if((t=getregnum(n))==-1) error("Illegal number register");
	return &numbers[t];
}

/*	============================================================
	Initialize the number register formats and values. Called by init.
*/

void
init_nums(void)
{
	int t;
	time_t clock;
	struct tm *timeptr;
	
	/* set all number formats to decimal, floating precision */
	
	for(t=0; t<arraysize(numbers); t++)
	{
		numbers[t].num_fmt='D';
		numbers[t].num_pad='0';
	}
	
	/* set special values */
	
	time(&clock);
	if(clock != (time_t)-1)
	{
		timeptr=localtime(&clock);
		
		setnumval('y', (timeptr->tm_year)+1900);	// year
		setnumval('m', timeptr->tm_mon+1);		// month
		numbers[getregnum('w')].num_fmt='w';
		setnumval('d', timeptr->tm_mday);		// day of month
		setnumval('w', timeptr->tm_wday);		// day of week
		numbers[getregnum('m')].num_fmt='m';
		setnumval('h', timeptr->tm_hour);		// hour
		setnumval('i', timeptr->tm_min);		// minute
		setnumval('s', timeptr->tm_sec);		// second
	}
	setnumval('p', 1);					// page #
	setnumval('n', 1);					// footnote #
}

/*	============================================================
	Set the value of a number register
*/ 

void
setnumval(int n, int v)
{
	pointnumreg(n)->num_value=v;
}


/*	============================================================
	Set a number format
	
   .F#x[=]format [,precision] [,padchar]
*/

void
setnumf(void)
{
	Num_struct *np;
	int precise,  pad;
	
	precise=pad=0;
	np=pointnumreg(getkey());  /* point to reg */
	skipequal();
	np->num_fmt=getnonb();  /* get a character option */
	
	if(comma_ck())
	{
		precise=getval(0);		/* get precision option */
		if(comma_ck())
		{
			pad=getval(0);		 /* get padding option */
		}
	}
	np->num_precise=precise;
	np->num_pad=pad;
}


/*	============================================================
	Set the number register (#0..#9 or #a..#z). The functions called
	here all abort if there is an error...
   	
	.#x[=]value[,increment]
*/

void
setnum(void)
{
	Num_struct *np;
	u_int inc=0;
	int c;
	
	np=pointnumreg(c=getkey());  /* point to num, error if illegal value */
	skipequal();
	np->num_value=getval(0);
	if(comma_ck()) inc=getval(0);
	np->num_inc=inc;
	
	if(f_toupper(c)=='P' && colactive()==0 && getlnum()<=1) setprintflag();
}

/*	============================================================
	Get the value in a number reg. Called from math
*/

u_int
getnumval(KCHAR n)
{
	return pointnumreg(n)->num_value;
}

/*	============================================================
	Expand a number. This is called by getkey() when an embedded
	number is encountered. We simply expand the number into a buffer
	and then push it back to the returned key buffer when it is
	pulled out later.
*/

void
expnum(void)
{
	char buff[100];
	KCHAR n=getkey();
	
	stringnum(n, buff, sizeof(buff)-2);
	returnstr(buff);
	
	pointnumreg(n)->num_value+=pointnumreg(n)->num_inc;  /* increment value */
}


/*	============================================================
	Expand a number register into a string. This is called by
	the footnote routines to get the footnote number in ascii.
   	
   	This routine is used by expnum() as well as the footnote routines.
*/


void
stringnum(KCHAR n, char *buff, int maxlen)
{
	static char *monthnames[]=
	{
		"January",	"February",		"March",	"April",
		"May",		"June",			"July",		"August",
		"September",	"October",		"November",	"December"
	};
	
	static char *daynames[]=
	{
		"Sunday",	"Monday",		"Tuesday",	"Wednesday",
		"Thursday",	"Friday",		"Saturday"
	};
	
	Num_struct *np;
	KCHAR c;
	u_int t;
	char precision, l;
	int value;
	char fmt;
	
	np=pointnumreg(n);	   /* point to number slot */
	
	fmt=np->num_fmt;		  /* get format		 */
	precision=np->num_precise;	  /* get precision	 */
	value=np->num_value;		  /* get value		 */
	
	switch(fmt)
	{
	
		/*	These options end with a 'goto EXIT' which skips the 
			precision routine which is done if a 'break' is done.
		*/
		
		case 'R':	/* roman numeral */
		case 'r':
			itoroman(buff, value);
			if(fmt=='r') lowstr(buff);
			break;
		
		case 'e':	/* english */
		case 'E':
			itoeng(buff, value);
			if(fmt=='E') *buff=f_toupper(*buff);
			goto EXIT;
		
		case 'a':	/* single ascii char */
		case 'A':
			if(value<' ' || value>'z') value='?';
			*buff++=value;
			if(value==control) *buff++=value;
			*buff='\0';
			goto EXIT;
		
		case 'm':	/* month name */
		case 'M':
			strcpy(buff, monthnames[(value-1) % 12]);
			goto EXIT;
		
		case '$':	/* number string */
			if(!numstringcount) error("Number string not defined");
			if(value) value--;
			if(istoken(c=numberstring[value % numstringcount]))
				 untokenize(buff, c);
			else
			{
				buff[0]=c;
				buff[1]='\0';
			}
			goto EXIT;
		
		case 'w':	/* weekday name */
		case 'W':
			strcpy(buff, daynames[value % 7]);
			goto EXIT;
		
		/* these options drop through to the precision routine */
		
		case 'H':	/* uppercase hex */
		case 'h':	/* lowercase hex */
			sprintf(buff, "%x", value);
			if(fmt=='H') upstr(buff);
			break;
		
		default:	/* decimal */
			sprintf(buff, "%u", value);
			break;
	}
	
	if(precision)	  /* pad number if precision specified */
	{
		l=strlen(buff);			  /* l==length of number */
		if(!(c=np->num_pad)) c=' ';	  /* c==padding character */
		if(l>precision)			 /* number too long,  */
		{
			*buff='\0';		   /* set to null		*/
			c='*';			   /* and pad with '*'s */
			l=0;
		}
		if(l<precision)			 /* number too short   */
		{
			t=precision-l;
			memmove(buff+t, buff, maxlen-t);
			memset(buff, c, t);
		}
	}

EXIT:;
}

/*	============================================================
	Convert a number to a roman numeral.
	
	Note: this is valid for 0..4999. The caller must ensure buffer
		  is at least 17 bytes long for MMMMDCCCLXXXVIII (4888).
*/

static void
itoroman(char *p, int n)
                 	   /* pointer to conversion buffer */
               		   /* number to convert			   */
{
	static char mchar[]={'M', 'D', 'C', 'L', 'X', 'V', 'I'	};
	static char schar[]={'C', 'C', 'X', 'X', 'I', 'I', ' '	};
	static short mvalue[]={1000, 500, 100, 50, 10, 5, 1	};
	static short svalue[]={100,  100, 10,  10, 1,  1, 0	};
	
	short i;
	
	if(n==0) strcpy(p, "0");
	
	else if(n>4999) strcpy(p, "*****");
	
	else
	{
		for(i=0; i<7; i++)
		{
			while(n>=mvalue[i])
			{
				*p++=mchar[i];
				n-=mvalue[i];
			}
			if(n>=mvalue[i]-svalue[i])
			{
				*p++=schar[i];
				*p++=mchar[i];
				n-=mvalue[i]-svalue[i];
			}
		}
		*p='\0';
	}
}


/*	============================================================
	Convert a number to english (123==one hundred twenty three
	
	See C CHEST, page 288. The longest possible string is 57 chars. Caller
	must ensure buffer is large enough.
*/

static int
itoeng(char *dest, int n)
{
	static char *onestab[]=
	{
		"zero",	   "one",	"two",		"three",	"four",
		"five",	   "six",	"seven",	"eight",	"nine",
		"ten",	   "eleven",	"twelve",	"thirteen",	"fourteen",
		"fifteen", "sixteen",	"seventeen",	"eighteen", 	"ninteen"
	};
	
	static char *tenstab[]=
	{
		"",		"ten",		"twenty",	"thirty",	"forty",
		"fifty",	"sixty",	"seventy",	"eighty",	"ninty"
	};
	
	*dest='\0';
	
	if(n==0)
	{
		strcpy(dest, "zero");
		goto EXIT;
	}
	
	if(n<0)
	{
		strcpy(dest, "minus ");
		n=-n;
	}
	
	if(n>=10000 && n<=20000)
	{
		strcat(dest, onestab[n/1000]);
		strcat(dest, (n%=1000) ? " thousand, " : " thousand");
	}
	else if(n>=1000)
	{
		if(n>=20000)
		{
			strcat(dest, (n>=30000) ? "thirty" : "twenty");
			if(n%=10000) strcat(dest, (n>=1000) ? "-" : " ");
		}
		if(n>=1000)
		{
			strcat(dest, onestab[n/1000]);
			strcat(dest, " ");
		}
		strcat(dest, (n%=1000) ? "thousand, " : "thousand");
	}
	if(n>=100)
	{
		strcat(dest, onestab[n/100]);
		strcat(dest, (n%=100) ? " hundred " : " hundred");
	}
	if(n>=20)
	{
		strcat(dest, tenstab[n/10]);
		if(n %= 10) strcat(dest, "-");
	}
	if(n) strcat(dest, onestab[n]);
	
EXIT:
	return strlen(dest);
}

/*	============================================================
	Define the number subsitution string
*/

void
setnumstr(void)
{
	KCHAR c;
	int m;
	
	char buff[NUMSTRMAX];
	
	/* first read the string--use read string routine... */
	
	skipequal();
	getdstring(buff, NUMSTRMAX-1, "number string");
	
	/*	Now expand/transfer string to number buffer. This is done
		so that embedded things are tokenized and each item will then
		take up one slot in the string. 
	*/
	
	m=addmacro(buff);
	numstringcount=0;
	while((c=getkey())!=CR)
	{
		if(istoken(c) && !is_dkey(c))
		{
			numstringcount=0;
			error("Only defined keys and ASCII permitted in number string");
		}
		numberstring[(int)numstringcount++]=c;
	}
	killmacro(m);
}

/*	============================================================
	This is needed by state to determine the size of the number 
	structure and its address.
*/

void
getnumstate(int *s, char **p)
{
	*s=sizeof(Num_struct)*10;
	*p=(char *) &numbers[0];
}
