/* strings.c 

   Set string sequences
   expand embedded string commands
*/

#include "vpt.h"

static void strval(char od, char cd);
static void upstring(char od, char cd);
static void lowstring(char od, char cd);
static void trimstr(char od, char cd);
static void v_index(char od, char cd);
static void lenstr(char od, char cd);
static void leftstr(char od, char cd);
static void midstr(char od, char cd);
static void rightstr(char od, char cd);
static void stringstr(char od, char cd);
static void commacheck(void);
static void endstrcmd(char d, char *buff);


/*	============================================================
	Copy a delimited string to a buffer.
   	
	If the string starts with a '[', '{', '<', or '(' it will be
	read until a ']', '}', '>', or ')' is encountered; otherwise
	the string will be read until a  CR is found.
*/

void
getstring(char *p, int maxsize, char *emsg)
        		/* where to stuff data		 */
            	/* max size of data			 */
           				/* error message if overflow */
{
	KCHAR odelim, cdelim, c;
	char buff[200];
	char *bp;
	char ncount=0;
	
	odelim=getnonb();
	
	if((cdelim=isdelm(odelim))==0)
	{
		returnkey(odelim);
		cdelim=CR;
		odelim=0;
	}
	
	for( ; ; )
	{
		if(!(c=getkey()))
			terminate("End of file encountered while looking for string delimiter");
		if(c==odelim) ncount++;	   /* track nested ()s  */
		if(c==cdelim)			  /* closing delim	   */
		{
			if(ncount==0) break;
			else ncount--;
		}
		if(c==INT_CR) c=CR;		   /* convert \CR to real CR */
		if(istoken(c))
		{
			untokenize(buff, c);
			bp=buff;
			while(*bp)
			{
				*p++=*bp++;
				maxsize--;
				if(maxsize<=0) break;
			}
		}
		else
		{
			*p++=c & 0xff;
			maxsize--;
			if(maxsize<=0) break;
		}
	}
	*p='\0';  /* null terminate string */
	
	if(cdelim==CR) returnkey(CR);
	
	if(maxsize<=0)
	{
		sprintf(buff,"String buffer overflow evaluating %s",emsg);
		terminate(buff);
	}
}

/*	this is the same as above, but it adds a CR and INT_DELIM to the
	end of the string. This is used to define headers, change bars, etc.
*/

void
getdstring(char *p, int maxsize, char *emsg)
        		/* where to stuff data		 */
            	/* max size of data			 */
           				/* error message if overflow */
{
	getstring(p, maxsize-4, emsg);
	p=strend(p);
	*(p++)=CR;
	*(p++)=INT_DELIM;
	*p='\0';
}


/* ============================================================
   Echo a string to stdout (EH=)
*/

void
echotext(void)
{
	char buff[300];
	
	skipequal();
	getstring(buff, sizeof(buff)-2, "echo string");
	fputs(buff, stdout);
	fflush(stdout);
}

/* ============================================================
   Embedded string commands (called from getkey())
*/

static char *semsg="string evaluation";

int
embeddedstr(KCHAR c1, KCHAR c2)
{
	
	static struct
	{
		char *name;
		void (*func)(char, char);
	}
	cmdtable[]=
	{
		{"left$",	leftstr,},
		{"len",		lenstr,},
		{"low$",	lowstring,},
		{"mid$",	midstr,},
		{"right$",	rightstr,},
		{"string$",	stringstr,},
		{"strchr",	v_index,},
		{"trim$",	trimstr,},
		{"val",		strval,},
		{"up$",		upstring,}
	};
	
	char cmdname[10];
	
	KCHAR c;
	int count, t;
	char opendelm, closedelm;
	
	cmdname[0]=c1;
	cmdname[1]=c2;
	count=2;
	for( ; ; )
	{
		c=getkey();
		if((closedelm=isdelm(c))!=0)
		{
			opendelm=c;
			cmdname[count]=0;
			break;
		}
		cmdname[count++]=c;
		if(!f_isprintable(c) || count>=sizeof(cmdname))
			error("unknown string command");
	}
	lowstr(cmdname);
	
	for(t=0; t<arraysize(cmdtable); t++)
	{
		if((strcmp(cmdtable[t].name, cmdname))==0)
		{
			(void)(*cmdtable[t].func)(opendelm, closedelm);
			return 0;
		}
	}
	error("Unknown embedded command");
	
	return 0;		/* unreachable */
}


/*	val(). 
	This strips an integer expression out of a string. In the process, any
	decimal portions are also stipped. The expression can be anywhere in the
	string and can consist of any number of digits and operators.
		  
	For example:
		val($100.44)						returns  100
		val(ralph)							returns  0
		val(101 trombones)					returns  101
		val(nancy is 12)					returns  12
		val(99+23-9)						returns  99+23-9
		val(123.44 + 22)					returns  123+22
		val[carrots cost $44.23/(18.3*99)]  returns  44/(18*99)
	   
	Caution, a successful parsing does not mean that a vaid expression exits. 
*/

static void
strval(char od, char cd)
{
	char buff[200];
	char *p, *e;
	
	char *valid="+-*/%()0123456789";
	
	returnkey(od);
	getstring(buff, sizeof(buff)-2, semsg);
	
	p=e=buff;
	while(!index(valid, *p) && *p) p++; /* skip leading junk */
	for( ; ; )
	{
		while(*p && index(valid, *p)) *e++=*p++;  /* copy valid chars */
		if(*p=='.')						  /* skip fractional part */
		{
			p++;
			while(f_isdigit((KCHAR)*p)) p++;
		}
		else if(*p==' ')				 /* skip white space */
		{
			while(*p==' ') p++;
		}
		else break;
	}
	
	if(e==buff) *e++='0';	/* nothing in string, make it "0" */
	*e='\0';
	returnstr(buff);
}

static void
upstring(char od, char cd)
{
	char buff[200];
	
	returnkey(od);
	getstring(buff, sizeof(buff)-2, semsg);
	upstr(buff);
	returnstr(buff);
}

static void
lowstring(char od, char cd)
{
	char buff[200];
	
	returnkey(od);
	getstring(buff, sizeof(buff)-2, semsg);
	lowstr(buff);
	returnstr(buff);
}

static void
trimstr(char od, char cd)
{
	char buff[200];
	char *p;
	
	returnkey(od);
	getstring(buff, sizeof(buff)-2, semsg);
	p=strend(buff)-1;
	while(*p==' ' && p>=buff) *(p--)='\0';
	returnstr(buff);
}

static void
v_index(char od, char cd)
{
	char buff[200];
	int target, pos;
	char *p;
	
	getstring(buff, sizeof(buff)-2, semsg);
	commacheck();
	
	target=getval(0);
	if(target<' ' || target>0xff)
		error("Target for STRCHR out of range");
	
	if((p=index(buff, target))!=NULL) pos=(p-buff)+1;
	else pos=0;
	sprintf(buff, "%d", pos);
	endstrcmd(cd, buff);
}

static void
lenstr(char od, char cd)
{
	char buff[200];
	int t;
	
	returnkey(od);
	getstring(buff, sizeof(buff)-2, semsg);
	t=strlen(buff);
	sprintf(buff, "%d", t);
	returnstr(buff);
}

static void
leftstr(char od, char cd)
{
	char buff[200];
	u_int len;
	
	getstring(buff, sizeof(buff)-2, semsg);
	commacheck();
	len=getval(0);
	if(strlen(buff)>len) buff[len]='\0';
	endstrcmd(cd, buff);
}

static void
midstr(char od, char cd)
{
	char buff[200];
	char *p;
	u_int start, len;
	
	getstring(buff, sizeof(buff)-2, semsg);
	commacheck();
	start=getval(0);
	if(!start) start=1;
	
	commacheck();
	len=getval(0);
	
	if(start>strlen(buff)) p=NULL;
	else
	{
		p=buff+start-1;
		if(len<strlen(p)) p[len]='\0';
	}
	endstrcmd(cd, p);
}

static void
rightstr(char od, char cd)
{
	char buff[200];
	u_int offset;
	char *p;
	
	getstring(buff, sizeof(buff)-2, semsg);
	commacheck();
	offset=getval(0);
	p=strend(buff)-offset;
	if(p<buff) p=buff;
	
	endstrcmd(cd, p);
}

static void
stringstr(char od, char cd)
{
	char buff[200];
	u_int count, fill;
	
	returnkey(getnonb());
	count=getval(0);
	commacheck();
	fill=getval(0);
	if(fill<' ' || fill>0xff) error("Illegal fill character for string$");
	if(count>sizeof(buff)-2) error("String$ overflow");
	
	memset(buff, fill, count);
	buff[count]='\0';
	endstrcmd(cd, buff);
}

static void
commacheck(void)
{
	if(!comma_ck()) error("Comma expected in string command");
	returnkey(getnonb());
}

static void
endstrcmd(char d, char *buff)
{
	if(getnonb()!=d) error("Wrong closing delimiter for string command");
	if(buff) returnstr(buff);
}
